"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var flushTime_1 = require("./utils/flushTime");
var sourcemap_codec_1 = require("sourcemap-codec");
var magic_string_1 = require("magic-string");
var object_1 = require("./utils/object");
var Module_1 = __importDefault(require("./Module"));
var index_1 = __importDefault(require("./finalisers/index"));
var getExportMode_1 = __importDefault(require("./utils/getExportMode"));
var getIndentString_1 = __importDefault(require("./utils/getIndentString"));
var promise_1 = require("./utils/promise");
var transformBundle_1 = __importDefault(require("./utils/transformBundle"));
var collapseSourcemaps_1 = __importDefault(require("./utils/collapseSourcemaps"));
var callIfFunction_1 = __importDefault(require("./utils/callIfFunction"));
var error_1 = __importDefault(require("./utils/error"));
var path_1 = require("./utils/path");
var ExternalModule_1 = __importDefault(require("./ExternalModule"));
var identifierHelpers_1 = require("./utils/identifierHelpers");
var LocalVariable_1 = __importDefault(require("./ast/variables/LocalVariable"));
var Chunk = /** @class */ (function () {
    function Chunk(graph, orderedModules) {
        var _this = this;
        this.graph = graph;
        this.orderedModules = orderedModules;
        this.exportedVariableNames = new Map();
        this.imports = [];
        this.exports = {};
        this.dependencies = undefined;
        this.entryModule = undefined;
        this.isEntryModuleFacade = orderedModules.length === 0;
        orderedModules.forEach(function (module) {
            if (module.isEntryPoint) {
                if (!_this.entryModule) {
                    _this.entryModule = module;
                    _this.isEntryModuleFacade = true;
                }
                else {
                    _this.isEntryModuleFacade = false;
                }
            }
            module.chunk = _this;
        });
    }
    Chunk.prototype.setId = function (id) {
        this.id = id;
        this.name = identifierHelpers_1.makeLegal(id);
    };
    // ensure that the module exports or reexports the given variable
    // we don't replace reexports with the direct reexport from the final module
    // as this might result in exposing an internal module which taints an entryModule chunk
    Chunk.prototype.ensureExport = function (module, variable, exportName) {
        // assert(module.chunk === this || module.isExternal);
        var safeExportName = this.exportedVariableNames.get(variable);
        if (safeExportName) {
            return safeExportName;
        }
        var i = 0;
        safeExportName = exportName;
        while (this.exports[safeExportName]) {
            safeExportName = exportName + '$' + ++i;
        }
        variable.exportName = safeExportName;
        this.exports[safeExportName] = { module: module, name: safeExportName, variable: variable };
        this.exportedVariableNames.set(variable, safeExportName);
        // if we've just exposed an export of a non-entry-point or had to use a safe name,
        // then note we are no longer an entry point chunk
        // we will then need an entry point facade if this is an entry point module
        if (this.isEntryModuleFacade && (!module.isEntryPoint || safeExportName !== exportName)) {
            this.isEntryModuleFacade = false;
        }
        return safeExportName;
    };
    Chunk.prototype.generateEntryExports = function (entryModule) {
        var _this = this;
        entryModule.getAllExports().forEach(function (exportName) {
            var traced = _this.traceExport(entryModule, exportName);
            var variable = traced.module.traceExport(traced.name);
            var tracedName;
            if (traced.module.chunk === _this || traced.module.isExternal) {
                tracedName = traced.name;
            }
            else {
                // if we exposed an export in another module ensure it is exported there
                tracedName = traced.module.chunk.ensureExport(traced.module, variable, traced.name);
            }
            _this.exports[exportName] = {
                module: traced.module,
                name: tracedName,
                variable: variable
            };
            _this.exportedVariableNames.set(variable, exportName);
        });
    };
    Chunk.prototype.collectDependencies = function (entryFacade) {
        var _this = this;
        if (entryFacade) {
            this.dependencies = [entryFacade.chunk];
            return;
        }
        this.dependencies = [];
        this.orderedModules.forEach(function (module) {
            module.dependencies.forEach(function (dep) {
                if (dep.chunk === _this) {
                    return;
                }
                var depModule;
                if (dep instanceof Module_1.default) {
                    depModule = dep.chunk;
                }
                else {
                    // unused pure external modules can be skipped
                    if (!dep.used && _this.graph.isPureExternalModule(dep.id)) {
                        return;
                    }
                    depModule = dep;
                }
                if (!_this.dependencies.some(function (dep) { return dep === depModule; })) {
                    _this.dependencies.push(depModule);
                }
            });
        });
        Object.keys(this.exports).forEach(function (exportName) {
            var expt = _this.exports[exportName];
            if (expt.module instanceof ExternalModule_1.default) {
                if (!_this.dependencies.some(function (dep) { return dep === expt.module; })) {
                    _this.dependencies.push(expt.module);
                }
            }
            else if (expt.module.chunk !== _this) {
                if (!_this.dependencies.some(function (dep) { return dep === expt.module.chunk; })) {
                    _this.dependencies.push(expt.module.chunk);
                }
            }
        });
    };
    Chunk.prototype.generateImports = function () {
        var _this = this;
        this.orderedModules.forEach(function (module) {
            Object.keys(module.imports).forEach(function (importName) {
                var declaration = module.imports[importName];
                _this.traceImport(declaration.module, declaration.name);
            });
        });
    };
    Chunk.prototype.populateImport = function (variable, tracedExport) {
        if (!variable.included) {
            return;
        }
        var exportName, importModule;
        // ensure that the variable is exported by the other chunk to this one
        if (tracedExport.module instanceof Module_1.default) {
            importModule = tracedExport.module.chunk;
            exportName = tracedExport.module.chunk.ensureExport(tracedExport.module, variable, tracedExport.name);
        }
        else {
            importModule = tracedExport.module;
            exportName = variable.name;
        }
        var impt = this.imports.find(function (impt) { return impt.module === importModule; });
        if (!impt) {
            this.imports.push(impt = { module: importModule, variables: [] });
        }
        // if we already import this variable skip
        if (impt.variables.some(function (v) { return v.module === tracedExport.module && v.variable === variable; })) {
            return;
        }
        impt.variables.push({
            module: tracedExport.module,
            variable: variable,
            name: exportName[0] === '*' ? '*' : exportName
        });
    };
    Chunk.prototype.getImportIds = function () {
        return this.dependencies.map(function (module) { return module.id; });
    };
    Chunk.prototype.getExportNames = function () {
        return Object.keys(this.exports);
    };
    Chunk.prototype.getJsonModules = function () {
        return this.orderedModules.map(function (module) { return module.toJSON(); });
    };
    Chunk.prototype.traceImport = function (module, exportName) {
        var _this = this;
        var tracedExport = this.traceExport(module, exportName);
        // ignore imports to modules already in this chunk
        if (!tracedExport || tracedExport.module.chunk === this) {
            return tracedExport;
        }
        var variable = tracedExport.module.traceExport(tracedExport.name);
        // namespace variable can indicate multiple imports
        if (tracedExport.name === '*') {
            Object.keys(variable.originals || variable.module.declarations).forEach(function (importName) {
                var original = (variable.originals || variable.module.declarations)[importName];
                _this.populateImport(original, {
                    name: importName,
                    module: tracedExport.module
                });
            });
            return tracedExport;
        }
        this.populateImport(variable, tracedExport);
        return tracedExport;
    };
    // trace a module export to its exposed chunk module export
    // either in this chunk or in another
    // we follow reexports if they are not entry points in the hope
    // that we can get an entry point reexport to reduce the chance of
    // tainting an entryModule chunk by exposing other unnecessary exports
    Chunk.prototype.traceExport = function (module, name) {
        if (name === '*') {
            return { name: name, module: module };
        }
        if (module instanceof ExternalModule_1.default) {
            return { name: name, module: module };
        }
        if (module.chunk !== this && module.isEntryPoint) {
            return { name: name, module: module };
        }
        var exportDeclaration = module.exports[name];
        if (exportDeclaration) {
            // if export binding is itself an import binding then continue tracing
            var importDeclaration = module.imports[exportDeclaration.localName];
            if (importDeclaration)
                return this.traceImport(importDeclaration.module, importDeclaration.name);
            return { name: name, module: module };
        }
        var reexportDeclaration = module.reexports[name];
        if (reexportDeclaration) {
            return this.traceExport(reexportDeclaration.module, reexportDeclaration.localName);
        }
        if (name === 'default') {
            return;
        }
        // external star exports
        if (name[0] === '*') {
            return { name: '*', module: this.graph.moduleById.get(name.substr(1)) };
        }
        // resolve known star exports
        for (var i = 0; i < module.exportAllModules.length; i++) {
            var exportAllModule = module.exportAllModules[i];
            // we have to ensure the right export all module
            if (exportAllModule.traceExport(name)) {
                return this.traceExport(exportAllModule, name);
            }
        }
    };
    Chunk.prototype.collectAddon = function (initialAddon, addonName, sep) {
        if (sep === void 0) { sep = '\n'; }
        return promise_1.runSequence([{ pluginName: 'rollup', source: initialAddon }]
            .concat(this.graph.plugins.map(function (plugin, idx) {
            return {
                pluginName: plugin.name || "Plugin at pos " + idx,
                source: plugin[addonName]
            };
        }))
            .map(function (addon) {
            addon.source = callIfFunction_1.default(addon.source);
            return addon;
        })
            .filter(function (addon) {
            return addon.source;
        })
            .map(function (_a) {
            var pluginName = _a.pluginName, source = _a.source;
            return Promise.resolve(source).catch(function (err) {
                error_1.default({
                    code: 'ADDON_ERROR',
                    message: "Could not retrieve " + addonName + ". Check configuration of " + pluginName + ".\n\tError Message: " + err.message
                });
            });
        })).then(function (addons) { return addons.filter(Boolean).join(sep); });
    };
    Chunk.prototype.setDynamicImportResolutions = function (_a) {
        var _this = this;
        var format = _a.format;
        var es = format === 'es';
        var dynamicImportMechanism;
        var hasDynamicImports = false;
        if (!es) {
            if (format === 'cjs') {
                dynamicImportMechanism = {
                    left: 'Promise.resolve(require(',
                    right: '))',
                    interopLeft: 'Promise.resolve({ default: require(',
                    interopRight: ') })'
                };
            }
            else if (format === 'amd') {
                dynamicImportMechanism = {
                    left: 'new Promise(function (resolve, reject) { require([',
                    right: '], resolve, reject) })',
                    interopLeft: 'new Promise(function (resolve, reject) { require([',
                    interopRight: '], function (m) { resolve({ default: m }) }, reject) })',
                };
            }
            else if (format === 'system') {
                dynamicImportMechanism = {
                    left: 'module.import(',
                    right: ')'
                };
            }
        }
        this.orderedModules.forEach(function (module) {
            module.dynamicImportResolutions.forEach(function (replacement, index) {
                var node = module.dynamicImports[index];
                hasDynamicImports = true;
                if (!replacement)
                    return;
                if (replacement instanceof Module_1.default) {
                    // if we have the module in the chunk, inline as Promise.resolve(namespace)
                    // ensuring that we create a namespace import of it as well
                    if (replacement.chunk === _this) {
                        node.setResolution(replacement.namespace(), false);
                        // for the module in another chunk, import that other chunk directly
                    }
                    else {
                        node.setResolution("\"" + replacement.chunk.id + "\"", false);
                    }
                    // external dynamic import resolution
                }
                else if (replacement instanceof ExternalModule_1.default) {
                    node.setResolution("\"" + replacement.id + "\"", true);
                    // AST Node -> source replacement
                }
                else {
                    node.setResolution(replacement, false);
                }
            });
        });
        if (hasDynamicImports)
            return dynamicImportMechanism;
    };
    Chunk.prototype.setIdentifierRenderResolutions = function (options) {
        var _this = this;
        var used = object_1.blank();
        var es = options.format === 'es';
        var system = options.format === 'system';
        // ensure no conflicts with globals
        Object.keys(this.graph.scope.variables).forEach(function (name) { return (used[name] = 1); });
        function getSafeName(name) {
            var safeName = name;
            while (used[safeName]) {
                safeName = name + "$" + used[name]++;
            }
            used[safeName] = 1;
            return safeName;
        }
        // reserved internal binding names for system format wiring
        if (system) {
            used['_setter'] = used['_starExcludes'] = used['_$p'] = 1;
        }
        var toDeshadow = new Set();
        if (!es) {
            this.dependencies.forEach(function (module) {
                if (module.isExternal) {
                    var safeName = getSafeName(module.name);
                    toDeshadow.add(safeName);
                    module.name = safeName;
                }
            });
        }
        this.imports.forEach(function (impt) {
            impt.variables.forEach(function (_a) {
                var name = _a.name, module = _a.module, variable = _a.variable;
                var safeName;
                if (module.isExternal) {
                    if (variable.name === '*') {
                        safeName = module.name;
                    }
                    else if (variable.name === 'default') {
                        if (module.exportsNamespace || !es && module.exportsNames) {
                            safeName = module.name + "__default";
                        }
                        else {
                            safeName = module.name;
                        }
                    }
                    else {
                        safeName = (es || system) ? variable.name : module.name + "." + name;
                    }
                    if (es || system) {
                        safeName = getSafeName(safeName);
                        toDeshadow.add(safeName);
                    }
                }
                else if (es || system) {
                    safeName = getSafeName(variable.name);
                }
                else {
                    safeName = module.chunk.name + "." + name;
                }
                variable.setSafeName(safeName);
            });
        });
        this.orderedModules.forEach(function (module) {
            object_1.forOwn(module.scope.variables, function (variable) {
                if (variable.isDefault && variable.referencesOriginal()) {
                    variable.setSafeName(null);
                    return;
                }
                if (!variable.isDefault || !variable.hasId) {
                    var safeName = void 0;
                    if (es || system || !variable.isReassigned || variable.isId) {
                        safeName = getSafeName(variable.name);
                    }
                    else {
                        var safeExportName = _this.exportedVariableNames.get(variable);
                        if (safeExportName) {
                            safeName = "exports." + safeExportName;
                        }
                        else {
                            safeName = getSafeName(variable.name);
                        }
                    }
                    variable.setSafeName(safeName);
                }
            });
            // deconflict reified namespaces
            var namespace = module.namespace();
            if (namespace.needsNamespaceBlock) {
                namespace.name = getSafeName(namespace.name);
            }
        });
        this.graph.scope.deshadow(toDeshadow, this.orderedModules.map(function (module) { return module.scope; }));
    };
    Chunk.prototype.getCheckReexportDeclarations = function () {
        var reexportDeclarations = {};
        for (var name_1 in this.exports) {
            var expt = this.exports[name_1];
            // skip local exports
            if (expt.module.chunk === this)
                continue;
            var depId = void 0;
            if (expt.module.isExternal) {
                depId = expt.module.id;
            }
            else {
                depId = expt.module.chunk.id;
            }
            var exportDeclaration = reexportDeclarations[depId] = reexportDeclarations[depId] || [];
            exportDeclaration.push({
                imported: expt.name,
                reexported: name_1[0] === '*' ? '*' : name_1
            });
        }
        return reexportDeclarations;
    };
    Chunk.prototype.getChunkDependencyDeclarations = function () {
        var _this = this;
        var reexportDeclarations = this.getCheckReexportDeclarations();
        var dependencies = [];
        this.dependencies.forEach(function (dep) {
            var importSpecifiers = _this.imports.find(function (impt) { return impt.module === dep; });
            var imports;
            if (importSpecifiers && importSpecifiers.variables.length) {
                imports = [];
                for (var i = 0; i < importSpecifiers.variables.length; i++) {
                    var impt = importSpecifiers.variables[i];
                    imports.push({
                        local: impt.variable.safeName || impt.variable.name,
                        imported: impt.name
                    });
                }
            }
            var reexports = reexportDeclarations[dep.id];
            var exportsNames, exportsNamespace, exportsDefault;
            if (dep.isExternal) {
                exportsNames = dep.exportsNames;
                exportsNamespace = dep.exportsNamespace;
                exportsDefault = 'default' in dep.declarations;
            }
            else {
                exportsNames = true;
                // we don't want any interop patterns to trigger
                exportsNamespace = false;
                exportsDefault = false;
            }
            dependencies.push({
                id: dep.id,
                name: dep.name,
                isChunk: !dep.isExternal,
                exportsNames: exportsNames,
                exportsNamespace: exportsNamespace,
                exportsDefault: exportsDefault,
                reexports: reexports,
                imports: imports
            });
        });
        return dependencies;
    };
    Chunk.prototype.getChunkExportDeclarations = function () {
        var exports = [];
        var _loop_1 = function (name_2) {
            var expt = this_1.exports[name_2];
            // skip external exports
            if (expt.module.chunk !== this_1)
                return "continue";
            // determine if a hoisted export (function)
            var hoisted = false;
            if (expt.variable instanceof LocalVariable_1.default) {
                expt.variable.declarations.forEach(function (decl) {
                    if (decl.type === "ExportDefaultDeclaration" /* ExportDefaultDeclaration */) {
                        if (decl.declaration.type === "FunctionDeclaration" /* FunctionDeclaration */)
                            hoisted = true;
                    }
                    else if (decl.parent.type === "FunctionDeclaration" /* FunctionDeclaration */) {
                        hoisted = true;
                    }
                });
            }
            exports.push({
                local: expt.variable.getName(),
                exported: name_2,
                hoisted: hoisted
            });
        };
        var this_1 = this;
        for (var name_2 in this.exports) {
            _loop_1(name_2);
        }
        return exports;
    };
    Chunk.prototype.getModuleDeclarations = function () {
        return {
            dependencies: this.getChunkDependencyDeclarations(),
            exports: this.getChunkExportDeclarations()
        };
    };
    Chunk.prototype.render = function (options) {
        var _this = this;
        return Promise.resolve()
            .then(function () {
            return Promise.all([
                _this.collectAddon(options.banner, 'banner'),
                _this.collectAddon(options.footer, 'footer'),
                _this.collectAddon(options.intro, 'intro', '\n\n'),
                _this.collectAddon(options.outro, 'outro', '\n\n')
            ]);
        })
            .then(function (_a) {
            var banner = _a[0], footer = _a[1], intro = _a[2], outro = _a[3];
            // Determine export mode - 'default', 'named', 'none'
            var exportMode = _this.isEntryModuleFacade ? getExportMode_1.default(_this, options) : 'named';
            var magicString = new magic_string_1.Bundle({ separator: '\n\n' });
            var usedModules = [];
            flushTime_1.timeStart('render modules');
            var renderOptions = {
                legacy: _this.graph.legacy,
                freeze: options.freeze !== false,
                systemBindings: options.format === 'system',
                importMechanism: _this.graph.dynamicImport && _this.setDynamicImportResolutions(options)
            };
            _this.setIdentifierRenderResolutions(options);
            _this.orderedModules.forEach(function (module) {
                var source = module.render(renderOptions);
                if (source.toString().length) {
                    magicString.addSource(source);
                    usedModules.push(module);
                }
            });
            if (!magicString.toString().trim() && _this.getExportNames().length === 0) {
                _this.graph.warn({
                    code: 'EMPTY_BUNDLE',
                    message: 'Generated an empty bundle'
                });
            }
            flushTime_1.timeEnd('render modules');
            var indentString = getIndentString_1.default(magicString, options);
            var finalise = index_1.default[options.format];
            if (!finalise) {
                error_1.default({
                    code: 'INVALID_OPTION',
                    message: "Invalid format: " + options.format + " - valid options are " + Object.keys(index_1.default).join(', ')
                });
            }
            flushTime_1.timeStart('render format');
            var getPath = _this.createGetPath(options);
            if (intro)
                intro += '\n\n';
            if (outro)
                outro = "\n\n" + outro;
            magicString = finalise(_this, magicString.trim(), // TODO TypeScript: Awaiting MagicString PR
            { exportMode: exportMode, getPath: getPath, indentString: indentString, intro: intro, outro: outro,
                dynamicImport: !!renderOptions.importMechanism }, options);
            flushTime_1.timeEnd('render format');
            if (banner)
                magicString.prepend(banner + '\n');
            if (footer)
                magicString.append('\n' + footer); // TODO TypeScript: Awaiting MagicString PR
            var prevCode = magicString.toString();
            var map = null;
            var bundleSourcemapChain = [];
            return transformBundle_1.default(prevCode, _this.graph.plugins, bundleSourcemapChain, options).then(function (code) {
                if (options.sourcemap) {
                    flushTime_1.timeStart('sourcemap');
                    var file = options.file ? options.sourcemapFile || options.file : _this.id;
                    if (file)
                        file = path_1.resolve(typeof process !== 'undefined' ? process.cwd() : '', file);
                    if (_this.graph.hasLoaders ||
                        _this.graph.plugins.find(function (plugin) { return Boolean(plugin.transform || plugin.transformBundle); })) {
                        map = magicString.generateMap({}); // TODO TypeScript: Awaiting missing version in SourceMap type
                        if (typeof map.mappings === 'string') {
                            map.mappings = sourcemap_codec_1.decode(map.mappings);
                        }
                        map = collapseSourcemaps_1.default(_this, file, map, usedModules, bundleSourcemapChain);
                    }
                    else {
                        map = magicString.generateMap({ file: file, includeContent: true }); // TODO TypeScript: Awaiting missing version in SourceMap type
                    }
                    map.sources = map.sources.map(path_1.normalize);
                    flushTime_1.timeEnd('sourcemap');
                }
                if (code[code.length - 1] !== '\n')
                    code += '\n';
                return { code: code, map: map }; // TODO TypeScript: Awaiting missing version in SourceMap type
            });
        });
    };
    Chunk.prototype.createGetPath = function (options) {
        var _this = this;
        var optionsPaths = options.paths;
        var getPath = typeof optionsPaths === 'function'
            ? function (id) { return optionsPaths(id, _this.id) || _this.graph.getPathRelativeToBaseDirname(id, _this.id); }
            : optionsPaths
                ? function (id) {
                    return optionsPaths.hasOwnProperty(id)
                        ? optionsPaths[id]
                        : _this.graph.getPathRelativeToBaseDirname(id, _this.id);
                }
                : function (id) { return _this.graph.getPathRelativeToBaseDirname(id, _this.id); };
        return getPath;
    };
    return Chunk;
}());
exports.default = Chunk;
